@@ -3,6 +3,7 @@
 from __future__ import division
 from decimal import Decimal
 from sql import Table
+from math import ceil, floor, log10
 
 from trytond.model import ModelView, ModelSQL, fields, Check
 from trytond.pyson import Eval
@@ -55,7 +56,7 @@ class Uom(ModelSQL, ModelView):
 
     @classmethod
     def __register__(cls, module_name):
-        cursor = Transaction().cursor
+        cursor = Transaction().connection.cursor()
         model_data = Table('ir_model_data')
         # Migration from 1.6: corrected misspelling of ounce (was once)
         cursor.execute(*model_data.update(
@@ -76,7 +77,7 @@ class Uom(ModelSQL, ModelView):
         cls._order.insert(0, ('name', 'ASC'))
         cls._error_messages.update({
                 'change_uom_rate_title': ('You cannot change Rate, Factor or '
-                    'Category on a Unit of Measure. '),
+                    'Category on a Unit of Measure.'),
                 'change_uom_rate': ('If the UOM is still not used, you can '
                     'delete it otherwise you can deactivate it '
                     'and create a new one.'),
@@ -134,13 +135,14 @@ class Uom(ModelSQL, ModelView):
             ('symbol',) + tuple(clause[1:]),
             ]
 
-    @staticmethod
-    def round(number, precision=1.0):
-        i, d = divmod(precision, 1)
-        base = round(number / precision)
-        # Instead of multiply by the decimal part, we must divide by the
-        # integer to avoid precision lost due to float
-        return (base * i) + ((base / (1 / d)) if d != 0 else 0)
+    def round(self, number):
+        return _round(self, number, func=round)
+
+    def ceil(self, number):
+        return _round(self, number, func=ceil)
+
+    def floor(self, number):
+        return _round(self, number, func=floor)
 
     @classmethod
     def validate(cls, uoms):
@@ -232,7 +234,7 @@ class Uom(ModelSQL, ModelView):
             amount = amount * to_uom.rate
 
         if round:
-            amount = cls.round(amount, to_uom.rounding)
+            amount = to_uom.round(amount)
 
         return amount
 
@@ -265,3 +267,24 @@ class Uom(ModelSQL, ModelView):
             new_price = new_price / Decimal(rate_format % to_uom.rate)
 
         return new_price
+
+
+def _round(uom, number, func=round):
+    precision = uom.rounding
+    # Convert precision into an integer greater than 1 to avoid precision lost.
+    # This works for most cases because rounding is often: n * 10**i
+    if precision < 1:
+        exp = -floor(log10(precision))
+        factor = 10 ** exp
+        number *= factor
+        precision *= factor
+    else:
+        factor = 1
+    # Divide by factor which is an integer to avoid precision lost due to
+    # multiplication by float < 1.
+    # example:
+    # >>> 3 * 0.1
+    # 0.30000000000000004
+    # >>> 3 / 10.
+    # 0.3
+    return func(number / precision) * precision / factor
